"""
Module to process Shimmer Plus data
"""
import random

import numpy as np
import pandas as pd


class ShimmerPlusReader:
    """
    Parses, deidentifies (timeshifts) and writes data
    generated by Shimmer Consensys GSR.

    Attributes
    ----------
    data : DataFrame
        DataFrame with the values that were read from the csv file

    units : pandas.Series
        a unit for each column in the dataframe
    """
    def __init__(self, path, delimiter=';'):
        """
        Reads the csv file generated by the Shimmer device and saves the
        parsed DataFrame.

        Parameters
        ----------
        path : str
            Path of the csv file.
        delimiter, optional : str
            delimiter used in the csv file
        """
        self.delimiter = delimiter
        self.data = pd.read_csv(path, sep=delimiter, skiprows=1)
        self.data = self.data.dropna(axis=1, how='all')
        self.units = self.data.iloc[0] # saving self.data units
        self.data = self.data.drop(0).astype(float)
        self.data['time'] = pd.to_datetime(
            self.data['Shimmer_40AC_Timestamp_Unix_CAL'],
            unit=self.units['Shimmer_40AC_Timestamp_Unix_CAL']).round('ms')
        self.data.drop(columns=['Shimmer_40AC_Timestamp_Unix_CAL'], inplace=True)
        self.data = self.data.drop_duplicates(subset='time')
        self.data.set_index('time', inplace=True, verify_integrity=True)
        self.data.sort_index(inplace=True)
        self.data['acc_mag'] = np.linalg.norm(self.data[['Shimmer_40AC_Accel_LN_X_CAL',
                                                         'Shimmer_40AC_Accel_LN_Y_CAL',
                                                         'Shimmer_40AC_Accel_LN_Z_CAL']],
                                              axis=1)
        self.units['acc_mag'] = self.units['Shimmer_40AC_Accel_LN_Z_CAL']

    def write(self, path):
        """
        Writes the DataFrame and units to the
        writing path in the same format as it was read.

        Parameters
        ----------
        path : str
            Path to writing csv file. Writing mode: 'w'.
        """
        write_df = self.data.drop(columns=['acc_mag'])
        write_df['Shimmer_40AC_Timestamp_Unix_CAL'] = (write_df.index.to_series().astype(int) / 1e6).round()
        units = self.units.drop('acc_mag').to_frame().transpose()
        write_df = pd.concat([units, write_df])
        with open(path, 'w') as f:
            f.write(f"\"sep={self.delimiter}\"\n")
            write_df.to_csv(f, index=None, sep=self.delimiter)

    def timeshift(self, shift='random'):
        """
        Timeshifts the data by shifting all time related columns.

        Parameters
        ----------
        shift : None/'random', pd.Timestamp or pd.Timedelta
            If shift is not specified, shifts the data by a random time interval
            between one month and two years to the past.

            If shift is a timdelta, shifts the data by that timedelta.

            If shift is a timestamp, shifts the data such that the earliest entry
            is at that timestamp and the remaining values keep the same time distance to the first entry.
        """
        if shift == 'random':
            one_month = pd.Timedelta('30 days').value
            two_years = pd.Timedelta('730 days').value
            random_timedelta = - pd.Timedelta(random.uniform(one_month, two_years)).round('ms')
            self.timeshift(random_timedelta)
        if isinstance(shift, pd.Timestamp):
            shift = shift.round('ms')
            timedeltas = self.data.index - self.data.index.min()
            self.data.index = shift + timedeltas
        if isinstance(shift, pd.Timedelta):
            self.data.index += shift.round('ms')
